==> klayout-0.14.tar.gz: klayout-0.14/src/layEditable.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layEditable.h"
#include "dbClipboard.h"
#include "tlAssert.h"

#include "PropertiesDialogImpl.h"

#include <algorithm>

namespace lay
{

// ----------------------------------------------------------------
//  Editable implementation

Editable::Editable (lay::Editables *editables)
  : mp_editables (editables)
{
  if (editables) {
    editables->m_editables.insert (this);
  }
}

Editable::~Editable ()
{
  //  erase the object from the table of enabled services
  if (mp_editables) {
    mp_editables->enable (this, false);
  }
}

// ----------------------------------------------------------------
//  Editables implementation

Editables::Editables (db::Manager *manager)
  : db::Object (manager), mp_properties_dialog (0)
{
  //  .. nothing yet ..
}

Editables::~Editables ()
{
  close_properties_dialog ();
}

void 
Editables::del ()
{
  if (has_selection ()) {

    close_properties_dialog ();

    //  begin the transaction
    tl_assert (! manager ()->transacting ());
    manager ()->transaction (tl::translate ("Delete")); 
    //  this dummy operation will update the screen:
    manager ()->queue (this, new db::Op ());

    for (iterator e = begin (); e != end (); ++e) {
      e->del ();
    }

    //  end the transaction
    manager ()->commit ();

  }
}

void 
Editables::cut ()
{
  if (has_selection ()) {

    close_properties_dialog ();

    //  begin the transaction
    tl_assert (! manager ()->transacting ());
    manager ()->transaction (tl::translate ("Cut")); 
    //  this dummy operation will update the screen:
    manager ()->queue (this, new db::Op ());

    db::Clipboard::instance ().clear ();
    for (iterator e = begin (); e != end (); ++e) {
      e->cut ();
    }

    //  end the transaction
    manager ()->commit ();

  }
}

void 
Editables::copy ()
{
  if (has_selection ()) {
    db::Clipboard::instance ().clear ();
    for (iterator e = begin (); e != end (); ++e) {
      e->copy ();
    }
  }
}

void 
Editables::paste ()
{
  if (! db::Clipboard::instance ().empty ()) {

    close_properties_dialog ();

    //  begin the transaction
    tl_assert (! manager ()->transacting ());
    manager ()->transaction (tl::translate ("Paste")); 
    //  this dummy operation will update the screen:
    manager ()->queue (this, new db::Op ());

    for (iterator e = begin (); e != end (); ++e) {
      e->paste ();
    }

    //  end the transaction
    manager ()->commit ();

  }
}

void 
Editables::enable (lay::Editable *obj, bool en)
{
  if (en) {
    m_enabled.insert (obj);
  } else {
    close_properties_dialog ();
    obj->select (db::DBox (), lay::Editable::Reset);  //  clear selection
    m_enabled.erase (obj);
  }
}

void 
Editables::clear_selection ()
{
  close_properties_dialog ();
  for (iterator e = begin (); e != end (); ++e) {
    e->select (db::DBox (), lay::Editable::Reset);  //  clear selection
  }
}

void 
Editables::select ()
{
  close_properties_dialog ();
  for (iterator e = begin (); e != end (); ++e) {
    if (m_enabled.find (&*e) != m_enabled.end ()) {
      e->select (db::DBox (), lay::Editable::Replace);  //  select "all"
    }
  }
}

void 
Editables::select (const db::DBox &box, lay::Editable::SelectionMode mode)
{
  close_properties_dialog ();
  for (iterator e = begin (); e != end (); ++e) {
    if (m_enabled.find (&*e) != m_enabled.end ()) {
      e->select (box, mode); 
    }
  }
}

template <class X, class Y>
struct first_of_pair_cmp_f 
{
  bool operator() (const std::pair<X,Y> &p1, const std::pair<X,Y> &p2) const
  {
    return p1.first < p2.first;
  }
};

void 
Editables::select (const db::DPoint &pt, lay::Editable::SelectionMode mode)
{
  close_properties_dialog ();

  //  in a first pass evaluate the point selection proximity value to select
  //  those plugins that are active.

  std::vector< std::pair <double, iterator> > plugins;

  for (iterator e = begin (); e != end (); ++e) {
    if (m_enabled.find (&*e) != m_enabled.end ()) {
      plugins.push_back (std::make_pair (e->click_proximity (pt), e));
    }
  }

  //  sort the plugins found by the proximity
  std::sort (plugins.begin (), plugins.end (), first_of_pair_cmp_f<double, iterator> ());

  //  and call select on those objects until the first one (with the least proximity)
  //  has something selected
  std::vector< std::pair<double, iterator> >::const_iterator pi = plugins.begin (); 
  for ( ; pi != plugins.end (); ++pi) {
    if (pi->second->select (db::DBox (pt, pt), mode)) {
      break;
    }
  }

  //  in replace mode clear the selections on the following plugins
  if (mode == lay::Editable::Replace && pi != plugins.end ()) {
    for (++pi ; pi != plugins.end (); ++pi) {
      pi->second->select (db::DBox (), lay::Editable::Reset);
    }
  }
}

bool 
Editables::begin_move (bool sel, const db::DPoint &p)
{
  close_properties_dialog ();
  if (sel) {
    for (iterator e = begin (); e != end (); ++e) {
      e->begin_move (true, p);
    }
    return true;
  } else {
    for (iterator e = begin (); e != end (); ++e) {
      if (e->begin_move (false, p)) {
        return true;
      }
    }
    return false;
  }
}

void 
Editables::move (const db::DPoint &p)
{
  for (iterator e = begin (); e != end (); ++e) {
    e->move (p);
  }
}

void 
Editables::move_transform (const db::DPoint &p, db::fixpoint_trans t)
{
  for (iterator e = begin (); e != end (); ++e) {
    e->move_transform (p, t);
  }
}

void 
Editables::end_move (const db::DPoint &p)
{
  for (iterator e = begin (); e != end (); ++e) {
    e->end_move (p);
  }
}

bool 
Editables::has_selection ()
{
  for (iterator e = begin (); e != end (); ++e) {
    if (e->has_selection ()) {
      return true;
    }
  }
  return false;
}

void 
Editables::edit_cancel ()
{
  for (iterator e = begin (); e != end (); ++e) {
    e->edit_cancel ();
  }
}

void
Editables::close_properties_dialog ()
{
  if (mp_properties_dialog) {
    delete mp_properties_dialog;
  }
  mp_properties_dialog = 0;
}

void
Editables::show_properties (QWidget *parent)
{
  close_properties_dialog ();
  mp_properties_dialog = new PropertiesDialogImpl (parent, manager (), this);
  mp_properties_dialog->show ();
}

}

==> klayout-0.15.tar.gz: klayout-0.15/src/layEditable.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layEditable.h"
#include "dbClipboard.h"
#include "tlAssert.h"

#include "PropertiesDialogImpl.h"

#include <algorithm>

namespace lay
{

// ----------------------------------------------------------------
//  Editable implementation

Editable::Editable (lay::Editables *editables)
  : mp_editables (editables)
{
  if (editables) {
    editables->m_editables.insert (this);
  }
}

Editable::~Editable ()
{
  //  erase the object from the table of enabled services
  if (mp_editables) {
    mp_editables->enable (this, false);
  }
}

// ----------------------------------------------------------------
//  Editables implementation

Editables::Editables (db::Manager *manager)
  : db::Object (manager), mp_properties_dialog (0), m_move_selection (false)
{
  //  .. nothing yet ..
}

Editables::~Editables ()
{
  close_properties_dialog ();
}

void 
Editables::del ()
{
  if (has_selection ()) {

    close_properties_dialog ();

    //  begin the transaction
    tl_assert (! manager ()->transacting ());
    manager ()->transaction (tl::translate ("Delete")); 
    //  this dummy operation will update the screen:
    manager ()->queue (this, new db::Op ());

    for (iterator e = begin (); e != end (); ++e) {
      e->del ();
    }

    //  end the transaction
    manager ()->commit ();

  }
}

void 
Editables::cut ()
{
  if (has_selection ()) {

    close_properties_dialog ();

    //  this dummy operation will update the screen:
    manager ()->queue (this, new db::Op ());

    db::Clipboard::instance ().clear ();
    for (iterator e = begin (); e != end (); ++e) {
      e->cut ();
    }

  }
}

void 
Editables::copy ()
{
  if (has_selection ()) {
    db::Clipboard::instance ().clear ();
    for (iterator e = begin (); e != end (); ++e) {
      e->copy ();
    }
  }
}

db::DBox 
Editables::selection_bbox ()
{
  db::DBox sel_bbox;
  for (iterator e = begin (); e != end (); ++e) {
    sel_bbox += e->selection_bbox ();
  }
  return sel_bbox;
}

void 
Editables::transform (const db::DCplxTrans &tr)
{
  if (has_selection ()) {

    try {

      tl_assert (! manager ()->transacting ());
      manager ()->transaction (tl::translate ("Transform")); 
      //  this dummy operation will update the screen:
      manager ()->queue (this, new db::Op ());

      for (iterator e = begin (); e != end (); ++e) {
        e->transform (tr);
      }

      //  end the transaction
      manager ()->commit ();

    } catch (...) {
      manager ()->clear ();
      throw;
    }

  }
}

void 
Editables::paste ()
{
  if (! db::Clipboard::instance ().empty ()) {

    close_properties_dialog ();

    //  this dummy operation will update the screen:
    if (manager ()->transacting ()) {
      manager ()->queue (this, new db::Op ());
    }

    for (iterator e = begin (); e != end (); ++e) {
      e->paste ();
    }

  }
}

void 
Editables::enable (lay::Editable *obj, bool en)
{
  if (en) {
    m_enabled.insert (obj);
  } else {
    close_properties_dialog ();
    obj->select (db::DBox (), lay::Editable::Reset);  //  clear selection
    m_enabled.erase (obj);
  }
}

void 
Editables::clear_selection ()
{
  close_properties_dialog ();
  for (iterator e = begin (); e != end (); ++e) {
    e->select (db::DBox (), lay::Editable::Reset);  //  clear selection
  }
}

void 
Editables::select ()
{
  close_properties_dialog ();
  for (iterator e = begin (); e != end (); ++e) {
    if (m_enabled.find (&*e) != m_enabled.end ()) {
      e->select (db::DBox (), lay::Editable::Replace);  //  select "all"
    }
  }
}

void 
Editables::select (const db::DBox &box, lay::Editable::SelectionMode mode)
{
  close_properties_dialog ();
  for (iterator e = begin (); e != end (); ++e) {
    if (m_enabled.find (&*e) != m_enabled.end ()) {
      e->select (box, mode); 
    }
  }
}

template <class X, class Y>
struct first_of_pair_cmp_f 
{
  bool operator() (const std::pair<X,Y> &p1, const std::pair<X,Y> &p2) const
  {
    return p1.first < p2.first;
  }
};

void 
Editables::select (const db::DPoint &pt, lay::Editable::SelectionMode mode)
{
  close_properties_dialog ();

  //  in a first pass evaluate the point selection proximity value to select
  //  those plugins that are active.

  std::vector< std::pair <double, iterator> > plugins;

  for (iterator e = begin (); e != end (); ++e) {
    if (m_enabled.find (&*e) != m_enabled.end ()) {
      plugins.push_back (std::make_pair (e->click_proximity (pt), e));
    }
  }

  //  sort the plugins found by the proximity
  std::sort (plugins.begin (), plugins.end (), first_of_pair_cmp_f<double, iterator> ());

  //  and call select on those objects until the first one (with the least proximity)
  //  has something selected
  std::vector< std::pair<double, iterator> >::const_iterator pi = plugins.begin (); 
  for ( ; pi != plugins.end (); ++pi) {
    if (pi->second->select (db::DBox (pt, pt), mode)) {
      break;
    }
  }

  //  in replace mode clear the selections on the following plugins
  if (mode == lay::Editable::Replace && pi != plugins.end ()) {
    for (++pi ; pi != plugins.end (); ++pi) {
      pi->second->select (db::DBox (), lay::Editable::Reset);
    }
  }
}

bool 
Editables::begin_move (const db::DPoint &p)
{
  close_properties_dialog ();
  m_move_selection = false;

  bool sel = has_selection ();
  if (sel) {

    //  if something is selected, issue a move operation on all editables
    for (iterator e = begin (); e != end (); ++e) {
      e->begin_move (true, p);
    }
    return true;

  } else {

    //  if something is selected, issue a move operation on all editables
    for (iterator e = begin (); e != end (); ++e) {
      if (e->begin_move (false, p)) {
        return true;
      }
    }

    //  nothing particular selected - select ourselves and start over
    select (p, Editable::Replace);

    //  now we assume we have a selection - try to begin_move on this.
    sel = has_selection ();
    if (sel) {
      m_move_selection = true;
      for (iterator e = begin (); e != end (); ++e) {
        e->begin_move (true, p);
      }
      return true;
    } else {
      return false;
    }

  }
}

void 
Editables::move (const db::DPoint &p)
{
  for (iterator e = begin (); e != end (); ++e) {
    e->move (p);
  }
}

void 
Editables::move_transform (const db::DPoint &p, db::fixpoint_trans t)
{
  for (iterator e = begin (); e != end (); ++e) {
    e->move_transform (p, t);
  }
}

void 
Editables::end_move (const db::DPoint &p)
{
  //  begin the transaction
  tl_assert (! manager ()->transacting ());
  manager ()->transaction (tl::translate ("Move")); 
  //  this dummy operation will update the screen:
  manager ()->queue (this, new db::Op ());

  for (iterator e = begin (); e != end (); ++e) {
    e->end_move (p);
  }

  //  end the transaction
  manager ()->commit ();

  //  clear the selection that was set previously
  if (m_move_selection) {
    clear_selection ();
  }
}

bool 
Editables::has_selection ()
{
  for (iterator e = begin (); e != end (); ++e) {
    if (e->has_selection ()) {
      return true;
    }
  }
  return false;
}

void 
Editables::edit_cancel ()
{
  for (iterator e = begin (); e != end (); ++e) {
    e->edit_cancel ();
  }
}

void
Editables::close_properties_dialog ()
{
  if (mp_properties_dialog) {
    delete mp_properties_dialog;
  }
  mp_properties_dialog = 0;
}

void
Editables::show_properties (QWidget *parent)
{
  close_properties_dialog ();
  mp_properties_dialog = new PropertiesDialogImpl (parent, manager (), this);
  mp_properties_dialog->show ();
}

}
